@@ -185,7 +185,7 @@ struct _grn_ctx_impl {
   /* output portion */
   struct {
     grn_obj *buf;
-    void (*func)(grn_ctx *, int, void *);
+    grn_recv_handler_func func;
     union {
       void *ptr;
       int fd;
@@ -222,6 +222,8 @@ struct _grn_ctx_impl {
   grn_array *values;        /* temporary objects */
   grn_pat *temporary_columns;
   grn_options *temporary_options;
+  grn_critical_section columns_cache_lock;
+  grn_hash *columns_cache;
   grn_hash *ios;        /* IOs */
   grn_com *com;
   unsigned int com_status;
@@ -240,6 +242,9 @@ struct _grn_ctx_impl {
   } temporary_open_spaces;
 };
 
+void grn_ctx_impl_columns_cache_delete(grn_ctx *ctx, grn_id table_id);
+void grn_ctx_impl_columns_cache_clear(grn_ctx *ctx);
+
 #ifdef __cplusplus
 }
 #endif
