@@ -158,76 +158,92 @@ grn_index_column_build_column(grn_ctx *ctx,
 grn_rc
 grn_index_column_build(grn_ctx *ctx, grn_obj *index_column)
 {
-  grn_obj *src, **cp, **col, *target;
+  grn_obj *src, *target;
   grn_id *s = DB_OBJ(index_column)->source;
+
   if (!(DB_OBJ(index_column)->source_size) || !s) { return ctx->rc; }
-  if ((src = grn_ctx_at(ctx, *s))) {
-    target = GRN_OBJ_TABLEP(src) ? src : grn_ctx_at(ctx, src->header.domain);
-    if (target) {
-      int i, ncol = DB_OBJ(index_column)->source_size / sizeof(grn_id);
-      grn_table_flags lexicon_flags;
-      grn_ii *ii = (grn_ii *)index_column;
-      grn_bool use_grn_ii_build;
-      grn_obj *tokenizer = NULL;
-      grn_table_get_info(ctx,
-                         ii->lexicon,
-                         &lexicon_flags,
-                         NULL,
-                         &tokenizer,
-                         NULL,
-                         NULL);
-      switch (lexicon_flags & GRN_OBJ_TABLE_TYPE_MASK) {
-      case GRN_OBJ_TABLE_PAT_KEY :
-      case GRN_OBJ_TABLE_DAT_KEY :
-        use_grn_ii_build = GRN_TRUE;
-        break;
-      default :
-        use_grn_ii_build = GRN_FALSE;
-        break;
-      }
-      const grn_column_flags flags = grn_ii_get_flags(ctx, ii);
-      if ((flags & GRN_OBJ_WITH_WEIGHT)) {
-        use_grn_ii_build = GRN_FALSE;
-      }
-      if ((flags & GRN_OBJ_WITH_POSITION) &&
-          (!tokenizer &&
-           !GRN_TYPE_IS_TEXT_FAMILY(ii->lexicon->header.domain))) {
-        /* TODO: Support offline index construction for WITH_POSITION
-         * index against UInt32 vector column. */
-        use_grn_ii_build = GRN_FALSE;
-      }
-      if ((col = GRN_MALLOC(ncol * sizeof(grn_obj *)))) {
-        for (cp = col, i = ncol; i; s++, cp++, i--) {
-          if (!(*cp = grn_ctx_at(ctx, *s))) {
-            ERR(GRN_INVALID_ARGUMENT, "source invalid, n=%d",i);
-            GRN_FREE(col);
-            return ctx->rc;
-          }
-          if (GRN_OBJ_TABLEP(grn_ctx_at(ctx, DB_OBJ(*cp)->range))) {
-            use_grn_ii_build = GRN_FALSE;
-          }
-        }
-        if (use_grn_ii_build) {
-          if (grn_index_chunk_split_enable) {
-            grn_ii_build2(ctx, ii, NULL);
-          } else {
-            grn_ii_build(ctx, ii, grn_index_sparsity);
-          }
-        } else {
-          for (i = 0; i < ncol; i++) {
-            grn_obj *column = col[i];
-            grn_index_column_build_column(ctx, index_column, target, column);
-          }
-        }
-        GRN_FREE(col);
-        grn_obj_touch(ctx, index_column, NULL);
-      }
+
+  src = grn_ctx_at(ctx, *s);
+  if (!src) {
+    ERR(GRN_INVALID_ARGUMENT, "invalid source");
+    return ctx->rc;
+  }
+
+  target = GRN_OBJ_TABLEP(src) ? src : grn_ctx_at(ctx, src->header.domain);
+  if (!target) {
+    ERR(GRN_INVALID_ARGUMENT, "invalid target");
+    return ctx->rc;
+  }
+
+  grn_table_flags lexicon_flags;
+  grn_ii *ii = (grn_ii *)index_column;
+  grn_bool use_grn_ii_build;
+  grn_obj *tokenizer = NULL;
+  grn_table_get_info(ctx,
+                     ii->lexicon,
+                     &lexicon_flags,
+                     NULL,
+                     &tokenizer,
+                     NULL,
+                     NULL);
+  switch (lexicon_flags & GRN_OBJ_TABLE_TYPE_MASK) {
+  case GRN_OBJ_TABLE_PAT_KEY :
+  case GRN_OBJ_TABLE_DAT_KEY :
+    use_grn_ii_build = GRN_TRUE;
+    break;
+  default :
+    use_grn_ii_build = GRN_FALSE;
+    break;
+  }
+  const grn_column_flags flags = grn_ii_get_flags(ctx, ii);
+  if ((flags & GRN_OBJ_WITH_WEIGHT)) {
+    use_grn_ii_build = GRN_FALSE;
+  }
+  if ((flags & GRN_OBJ_WITH_POSITION) &&
+      (!tokenizer &&
+       !GRN_TYPE_IS_TEXT_FAMILY(ii->lexicon->header.domain))) {
+    /* TODO: Support offline index construction for WITH_POSITION
+     * index against UInt32 vector column. */
+    use_grn_ii_build = GRN_FALSE;
+  }
+
+  int n_columns = DB_OBJ(index_column)->source_size / sizeof(grn_id);
+  grn_obj **columns = GRN_MALLOC(n_columns * sizeof(grn_obj *));
+  if (!columns) {
+    return ctx->rc;
+  }
+
+  grn_obj **cp;
+  int i;
+  for (cp = columns, i = n_columns;
+       i > 0;
+       s++, cp++, i--) {
+    if (!(*cp = grn_ctx_at(ctx, *s))) {
+      ERR(GRN_INVALID_ARGUMENT, "source invalid, n=%d", i);
+      GRN_FREE(columns);
+      return ctx->rc;
+    }
+    if (GRN_OBJ_TABLEP(grn_ctx_at(ctx, DB_OBJ(*cp)->range))) {
+      use_grn_ii_build = GRN_FALSE;
+    }
+  }
+  grn_obj_set_visibility(ctx, index_column, false);
+  if (use_grn_ii_build) {
+    if (grn_index_chunk_split_enable) {
+      grn_ii_build2(ctx, ii, NULL);
     } else {
-      ERR(GRN_INVALID_ARGUMENT, "invalid target");
+      grn_ii_build(ctx, ii, grn_index_sparsity);
     }
   } else {
-    ERR(GRN_INVALID_ARGUMENT, "invalid source");
+    for (int i = 0; i < n_columns; i++) {
+      grn_obj *column = columns[i];
+      grn_index_column_build_column(ctx, index_column, target, column);
+    }
   }
+  grn_obj_set_visibility(ctx, index_column, true);
+  GRN_FREE(columns);
+  grn_obj_touch(ctx, index_column, NULL);
+
   return ctx->rc;
 }
 
@@ -1290,3 +1306,131 @@ exit :
 
   GRN_API_RETURN(ctx->rc);
 }
+
+grn_obj *
+grn_index_column_get_tokenizer(grn_ctx *ctx, grn_obj *index_column)
+{
+  grn_obj *tokenizer;
+  grn_obj *lexicon;
+
+  lexicon = grn_ctx_at(ctx, index_column->header.domain);
+  if (!lexicon) {
+    return NULL;
+  }
+
+  grn_table_get_info(ctx, lexicon, NULL, NULL, &tokenizer, NULL, NULL);
+  return tokenizer;
+}
+
+static bool
+grn_index_column_is_usable_equal(grn_ctx *ctx,
+                                 grn_obj *index_column,
+                                 grn_operator op)
+{
+  grn_obj *tokenizer;
+  tokenizer = grn_index_column_get_tokenizer(ctx, index_column);
+  if (!tokenizer) {
+    return true;
+  }
+  /* TODO: Improve this. Which tokenizer can be used with equal operation?
+     Do we need metadata in tokenizer? */
+  if (DB_OBJ(tokenizer)->id == GRN_DB_DELIMIT) {
+    return true;
+  }
+  return false;
+}
+
+static bool
+grn_index_column_is_usable_match(grn_ctx *ctx,
+                                 grn_obj *index_column,
+                                 grn_operator op)
+{
+  return true;
+}
+
+static bool
+grn_index_column_is_usable_range(grn_ctx *ctx,
+                                 grn_obj *index_column,
+                                 grn_operator op)
+{
+  grn_obj *tokenizer;
+  grn_obj *lexicon;
+
+  lexicon = grn_ctx_at(ctx, index_column->header.domain);
+  if (!lexicon) {
+    return false;
+  }
+
+  switch (lexicon->header.type) {
+  case GRN_TABLE_PAT_KEY :
+  case GRN_TABLE_DAT_KEY :
+    break;
+  default :
+    return false;
+    break;
+  }
+
+  grn_table_get_info(ctx, lexicon, NULL, NULL, &tokenizer, NULL, NULL);
+  return tokenizer == NULL;
+}
+
+static bool
+grn_index_column_is_usable_regexp(grn_ctx *ctx,
+                                  grn_obj *index_column,
+                                  grn_operator op)
+{
+  grn_obj *tokenizer;
+  tokenizer = grn_index_column_get_tokenizer(ctx, index_column);
+  if (!tokenizer) {
+    return false;
+  }
+  return tokenizer == grn_ctx_get(ctx, "TokenRegexp", -1);
+}
+
+bool
+grn_index_column_is_usable(grn_ctx *ctx,
+                           grn_obj *index_column,
+                           grn_operator op)
+{
+  GRN_API_ENTER;
+
+  if (!grn_obj_is_index_column(ctx, index_column)) {
+    GRN_API_RETURN(false);
+  }
+
+  if (!grn_obj_is_visible(ctx, index_column)) {
+    GRN_API_RETURN(false);
+  }
+
+  bool usable = false;
+  switch (op) {
+  case GRN_OP_EQUAL :
+  case GRN_OP_NOT_EQUAL :
+    usable = grn_index_column_is_usable_equal(ctx, index_column, op);
+    break;
+  case GRN_OP_MATCH :
+  case GRN_OP_NEAR :
+  case GRN_OP_NEAR2 :
+  case GRN_OP_SIMILAR :
+  case GRN_OP_PREFIX :
+  case GRN_OP_SUFFIX :
+  case GRN_OP_FUZZY :
+  case GRN_OP_QUORUM :
+    usable = grn_index_column_is_usable_match(ctx, index_column, op);
+    break;
+  case GRN_OP_LESS :
+  case GRN_OP_GREATER :
+  case GRN_OP_LESS_EQUAL :
+  case GRN_OP_GREATER_EQUAL :
+  case GRN_OP_CALL :
+    usable = grn_index_column_is_usable_range(ctx, index_column, op);
+    break;
+  case GRN_OP_REGEXP :
+    usable = grn_index_column_is_usable_regexp(ctx, index_column, op);
+    break;
+  default :
+    break;
+  }
+
+  GRN_API_RETURN(usable);
+}
